@@ -5,7 +5,7 @@
 # This is free software; you can redistribute it and/or modify it under
 # the same terms as the Perl 5 programming language system itself.
 #
-# 20230630 raf <raf@raf.org>
+# 20230706 raf <raf@raf.org>
 
 package Lingua::NameUtils;
 use 5.014;
@@ -13,7 +13,7 @@ use strict;
 use warnings;
 use utf8;
 
-our $VERSION = '1.000';
+our $VERSION = '1.001';
 
 use Exporter;
 our @ISA = ('Exporter');
@@ -22,14 +22,14 @@ our @EXPORT = ();
 our @EXPORT_OK =
 qw(
 	namecase gnamecase fnamecase namecase_exception
-	namesplit nameparts namesplit_exception
+	namesplit nameparts namesplit_exception namejoin
 	nametrim normalize
 );
 our %EXPORT_TAGS =
 (
 	all => [@EXPORT_OK],
 	case => [qw(namecase gnamecase fnamecase namecase_exception normalize)],
-	split => [qw(namesplit nameparts namesplit_exception normalize)]
+	split => [qw(namesplit nameparts namesplit_exception namejoin normalize)]
 );
 
 # Like fc() but "folds" apostrophe-like and hyphen-like characters as well
@@ -41,15 +41,14 @@ sub kc { return $myfc->(shift) =~ s/$apostrophe/'/gr =~ s/$hyphen/-/gr }
 
 # Builtin namecase exceptions (Mostly gathered by Michael R. Davis (MRDVT))
 
-my %namecase_exceptions = map { kc($_) => $_ }
+my @namecase_exceptions =
 (qw(
-
 	MacAlister MacAlpin MacAlpine MacArthur MacAuley MacBain MacBean
 	MacCallum MacColl MacDomhnaill MacDonald MacDonell MacDonnell MacDougall
 	MacDowall MacDuff MacEvan MacEwen MacFarlane MacFie MacGill MacGillivray
 	MacGregor MacInnes MacIntosh MacIntyre MacIver MacKay MacKenzie
 	MacKinlay MacKinnon MacKintosh MacLachlan MacLaine MacLaren MacLaurin
-	MacLea MacLean MacLellan MacLennan MacLeod MacMillan MacNab MacNaughton
+	MacLea MacLean MacLeay MacLellan MacLennan MacLeod MacMillan MacNab MacNaughton
 	MacNeacail MacNeil MacNeill MacNicol MacO'Shannaig MacPhee MacPherson
 	MacQuarrie MacQueen MacRae MacTavish MacThomas
 
@@ -115,9 +114,10 @@ my %namecase_exceptions = map { kc($_) => $_ }
 	VanLeer VanLiere VanLuven VanMeter VanOlinda VanOrmer VanPelt VanSchaick
 	VanSciver VanScoy VanScoyk VanSickle VanTassel VanTuyl VanValkenburg
 	VanVolkenburgh VanWinkle VanWysenberghe VanZandt VenDerWeyer VonCannon
-
 ));
 
+my %namecase_exceptions;
+
 # Capitalization exceptions for full names used by namecase().
 # Include both forms: "given_names family_name" and "family_name, given_names".
 
@@ -128,6 +128,38 @@ my %namecase_exceptions_full;
 
 my %fnamecase_exceptions_full;
 
+# Non-individual case exception replacements need a long regex that needs to
+# be constructed, and can change. But probably not often. And sometimes a
+# lot at once. Regex construction is cached lazily when needed for use.
+
+my $need_case_update = 1;
+my $namecase_exceptions_re;
+
+# Name affixes that start a multi-word family name
+
+my %split_starter;
+my $split_starter_re;
+my @split_starter =
+(qw(
+	de de’ de' del dels dela della delle dal dalla degli di da du do dos das
+	le la li lo y i
+	van von zu der ter den af av til
+	el al ibn bin ben bat bint binti binte mibeit mimishpachat
+	of o ó ni ní mac nic ua bean ui uí mhic ap ab ferch verch
+	san santa santos st st. ste ste.
+	ka te
+));
+
+# Data for regexes that are affected by normalization
+
+my @irish_o = qw(O Ó);
+my @irish_vowel = qw(a e i o u á é í ó ú);
+my @irish_post_bean = qw(Uí Ui Mhic);
+
+my $irish_o_re = "(?:@{[join '|', @irish_o]})";
+my $irish_vowel_re = "(?:@{[join '|', @irish_vowel]})";
+my $irish_post_bean_re = "(?:@{[join '|', @irish_post_bean]})";
+
 # Family names that appear first (Chinese, Korean, Vietnamese).
 # When romanized, these family names can appear first or last.
 # This is only possible when there are only hundreds of names.
@@ -700,28 +732,6 @@ Trang Trầm Trâu Trì Triệu Trịnh Từ Tư_Mã Tưởng Úc Ứng Vạn V
 Xầm Xế Yên Yến
 ));
 
-my %split_starter;
-my $split_starter_re;
-my $kcben = kc('ben');
-my $kcbean = kc('bean');
-my @split_starter =
-(qw(
-	de de’ de' del dels dela della delle dal dalla degli di da du do dos das
-	le la li lo y i
-	van von zu der ter den af av til
-	el al ibn bin ben bat bint binti binte mibeit mimishpachat
-	of o ó ni ní mac nic ua bean ui uí mhic ap ab ferch verch
-	san santa santos st st. ste ste.
-	ka te
-));
-
-# Non-individual case exception replacements need a long regex that needs to
-# be constructed, and can change. But probably not often. And sometimes a
-# lot at once. Regex construction is cached lazily when needed for use.
-
-my $need_case_update = 1;
-my $keys_namecase_exceptions;
-
 # Return the supplied full/given/family name with the case fixed
 
 sub namecase
@@ -737,9 +747,11 @@ sub namecase
 	$name = lc(nametrim($name));
 	$name =~ s/\b(\w)/\U$1/g;
 
-	# Lowercase after apostrophes that follow more than 1 letter (e.g. Oso'ese but not O'Brien)
+	# Lowercase after apostrophes that follow more than one letter (e.g. Oso'ese but not O'Brien)
 	$name =~ s/(?<=\w{2}|$apostrophe\w)($apostrophe\w)/\L$1/g;
-	$name =~ s/(?<=\b[YS])($apostrophe\w)/\L$1/g;
+	# Lowercase after apostrophes that follow one letter that isn't O, V or D
+	# (e.g. T'ang, but not O'Brien or d'Iapico or v'Rachel)
+	$name =~ s/(?<=\b[^ODV])($apostrophe\w)/\L$1/g;
 
 	# Uppercase after "Mc" and "Fitz" ("Mac" is done selectively with built-in exceptions)
 	$name =~ s/\b(Mc|Fitz)(\w)/$1\U$2/g;
@@ -768,11 +780,11 @@ sub namecase
 		# Note1: "ben" is only detected when unambiguous
 
 		$name =~ s/\b(d$apostrophe|(?:de(?: la|$apostrophe)?|del|dela|dels|della|delle|dal|dalla|degli|di|du|da|do|dos|das|y|i|e|von und zu|von|zu|van het|van|der|ter|den|tot|af|av|til|ap|ab|ferch|verch|ibn|bin|bint|binti|binte|bat|mibeit|mimishpachat|ka|of)\s)/\L$1/ig;
-		$name =~ s/\b(dall|dell)($apostrophe)(\w)/\L$1\E$2\U$3/ig;  # Italian: dall'Agnese
-		$name =~ s/((?:^|\s)${apostrophe})([st])(\w)/$1\L$2\U$3/ig; # Dutch: 'sGravesande
-		$name =~ s/\b([OÓ] )(h)([aeiouáéíóú])/$1\L$2\U$3/ig;        # Irish: Ó hUiginn
-		$name =~ s/\b(el|al|ut|ha)(?=$hyphen)/\L$1/ig;              # Arabic/Hebrew: el- al- ut- ha-
-		$name =~ s/\b(v)(?=$apostrophe)/\L$1/ig;                    # Hebrew: v'Rachel
+		$name =~ s/\b(dall|dell)($apostrophe)(\w)/\L$1\E$2\U$3/ig;     # Italian: dall'Agnese
+		$name =~ s/((?:^|\s)${apostrophe})([st])(\w)/$1\L$2\U$3/ig;    # Dutch: 'sGravesande
+		$name =~ s/\b($irish_o_re )(h)($irish_vowel_re)/$1\L$2\U$3/ig; # Irish: Ó hUiginn
+		$name =~ s/\b(el|al|ut|ha)(?=$hyphen)/\L$1/ig;                 # Arabic/Hebrew: el- al- ut- ha-
+		$name =~ s/\b(v)(?=$apostrophe)/\L$1/ig;                       # Hebrew: v'Rachel
 		$name =~ s/^(ben\s)/\L$1/i if $mode eq 'family' || index($name, ',') != -1; # Hebrew: ben if family
 		$name =~ s/(?<=\s)\b(ben)\b(?=\s)/\L$1/i if $name =~ / v$apostrophe| ha$hyphen(?:Kohein|Levi|Rav)\b/; # Hebrew: ben if v' or ha-
 
@@ -803,11 +815,12 @@ sub namecase
 
 	if ($need_case_update)
 	{
-		$keys_namecase_exceptions = join '|', keys %namecase_exceptions;
+		%namecase_exceptions = map { kc($_) => $_ } @namecase_exceptions unless %namecase_exceptions;
+		$namecase_exceptions_re = join '|', keys %namecase_exceptions;
 		$need_case_update = 0;
 	}
 
-	$name =~ s/\b($keys_namecase_exceptions)\b/$namecase_exceptions{kc($1)}/ieg;
+	$name =~ s/\b($namecase_exceptions_re)\b/$namecase_exceptions{kc($1)}/ieg;
 
 	return $name;
 }
@@ -921,7 +934,7 @@ sub namesplit
 	if (!scalar %family_names_ck_roman)
 	{
 		%family_names_ck_roman =
-			map { kc($_) => 1 }
+			map { (kc($_) =~ s/'/$apostrophe/r) => 1 }
 			grep { !/^$split_starter_re$/ }
 			@family_names_chinese_roman,
 			@family_names_korean_roman;
@@ -951,8 +964,8 @@ sub namesplit
 	{
 		my $kcstarter = kc($words[$i]);
 		next unless exists $split_starter{$kcstarter};
-		next if $kcstarter eq $kcben && $name !~ / v$apostrophe| ha$hyphen(?:Kohein|Levi|Rav)\b/i; # Hebrew
-		next if $kcstarter eq $kcbean && $name !~ /\b$kcbean (?:Uí|Ui|Mhic)\b/i; # Irish
+		next if $kcstarter eq 'ben' && $name !~ / v$apostrophe| ha$hyphen(?:Kohein|Levi|Rav)\b/i; # Hebrew
+		next if $kcstarter eq 'bean' && $name !~ /\bbean $irish_post_bean_re\b/i; # Irish
 		next if $i == $#words;
 
 		--$i if $i > 1 && $kcstarter =~ /^[yi]$/i; # Spanish/Catalan
@@ -991,7 +1004,6 @@ sub namesplit
 			$ja_loaded = 1;
 		}
 
-		local $SIG{__WARN__} = sub {}; # Avoid warnings for 々 (kanji repeater symbol)
 		($f, $g) = Lingua::JA::Name::Splitter::split_kanji_name($name);
 		return "$f, $g";
 	}
@@ -1048,6 +1060,19 @@ sub nameparts
 	return split /, ?/, namesplit($name), 2;
 }
 
+# Format a full name in Eastern or Western name order as appropriate
+
+sub namejoin
+{
+	my ($f, $g) = @_;
+
+	return $f if !defined $g;
+	return $g if !defined $f;
+
+	return "$f$g" if "$f$g" =~ /^[\p{Han}\p{Hangul}\p{Hiragana}\p{Katakana}]+$/;
+	return "$g $f";
+}
+
 # Trim the supplied name
 
 sub nametrim
@@ -1076,11 +1101,15 @@ sub normalize
 	%namecase_exceptions = map { $func->($_) => $func->($namecase_exceptions{$_}) } keys %namecase_exceptions;
 	%namecase_exceptions_full = map { $func->($_) => $func->($namecase_exceptions_full{$_}) } keys %namecase_exceptions_full;
 	%fnamecase_exceptions_full = map { $func->($_) => $func->($fnamecase_exceptions_full{$_}) } keys %fnamecase_exceptions_full;
-	$keys_namecase_exceptions = $func->($keys_namecase_exceptions) if defined $keys_namecase_exceptions;
+	$namecase_exceptions_re = $func->($namecase_exceptions_re) if defined $namecase_exceptions_re;
 	%namesplit_exceptions = map { $func->($_) => $func->($namesplit_exceptions{$_}) } keys %namesplit_exceptions;
 	%split_starter = map { $func->($_) => 1 } keys %split_starter;
-	$kcben = $func->($kcben);
-	$kcbean = $func->($kcbean);
+	@irish_o = map { $func->($_) } @irish_o;
+	$irish_o_re = '(?:' . join('|', @irish_o) . ')';
+	@irish_vowel = map { $func->($_) } @irish_vowel;
+	$irish_vowel_re = '(?:' . join('|', @irish_vowel) . ')';
+	@irish_post_bean = map { $func->($_) } @irish_post_bean;
+	$irish_post_bean_re = '(?:' . join('|', @irish_post_bean) . ')';
 
 	%family_names_ck = map { $func->($_) => 1 } keys %family_names_ck;
 	$family_names_ck_re = qr/(?:@{[join '|', keys %family_names_ck]})/;
